/*
 * example sketch to list a directory and create/read/write a file
 * on an mmc, using the arduino wrapper function library for fatfs
 *
 * (c) 2010 david sirkin sirkin@stanford.edu
 */

/* import the avr's power library for the clock_prescale_set macro. */

#include <avr/power.h>

/* import the full fatfs library for access to functions & defines. */

#include <diskio.h>
#include <fatfs.h>
#include <ff.h>
#include <ffconf.h>
#include <integer.h>

#define cs_pin      14             // chip select pin on the arduino
#define read_buffer 32             // size (in bytes) of read buffer 

/*
 * initialize the processor speed, setup the fatfs sd card (or, mms)
 * filesystem and begin serial communication.
 */
void setup() {
  // the avr atmega32u4 processor is only rated to 8mhz (rather than
  // 16mhz) when powered at 3.3v (rather than 5v).

  clock_prescale_set(clock_div_2); // fosc 8mhz as teensy is at 3.3v
  FatFs.begin(cs_pin);             // initialize FatFS library calls
  Serial.begin(115200);            // initialize the serial terminal
}

/*
 * read in buffer 'bytes' of 'read_buffer' size from file 'file' and
 * send the output to the serial terminal. but rewind the read/write
 * pointer back to 0, otherwise consecutive reads would get garbage.
 */
void file_read_and_print_to_serial (char * file) {
  char bytes[read_buffer + 1];     // buffer of bytes read plus '\0'
  unsigned int bytes_read;         // number of bytes read each call
  unsigned long bytes_count = 0;   // running count of bytes read
  
  Serial.print("Reading from file "); Serial.println(file);
  
  FatFs.lseek(0);                  // set file r/w pointer back to 0
  do {
    FatFs.read(bytes, read_buffer, &bytes_read);
    bytes_count += bytes_read;
    bytes[bytes_read] = '\0';
    
    Serial.print(bytes);
    delay(100);
  }
  while (bytes_read == read_buffer);
  
  Serial.print("\nRead "); Serial.print(bytes_count); Serial.println(" bytes.");
}

/*
 * write out buffer 'bytes_to_write' to file 'file' and send written
 *  bytes to the serial terminal. don't rewind the read/write pointer
 * so that consecutive writes will append to the end of the file.
 */
void file_write (char * bytes_to_write, char * file) {
  // make sure that the 'bytes_written' data type is large enough to
  // represent the 'bytes_to_write' buffer without overflowing.
  
  unsigned int bytes_written;
  
  Serial.print("Writing to file "); Serial.println(file);

  FatFs.write(bytes_to_write, strlen(bytes_to_write), &bytes_written);
    
  Serial.print("Wrote "); Serial.print(bytes_written); Serial.println(" bytes.");
}

/*
 * write out buffer 'bytes_to_write' to file 'file' and send written
 * list the files in the root '/' directory to the serial terminal.
 */
void dir_list (void) {
  FILINFO file_info;            // information on file size and type
  unsigned int files_found = 0; // count of files found in directory
  
  FatFs.opendir();
  do {
    FatFs.readdir(&file_info);
    
    if ((file_info.fattrib & AM_DIR) != AM_DIR) {
      Serial.println(file_info.fname);
      files_found++;
      delay(100);
    }
  }
  while (file_info.fname[0] != 0);
  
  Serial.print(files_found); Serial.println(" files found.");
}

/*
 * create file test2.txt if it doesn't already exist (truncate it if
 * it does) and write some text into it. read back what we wrote, to
 * confirm that the write worked. then list the new directory files.
 */
void loop() {
  // visit: http://elm-chan.org/fsw/ff/00index_e.html for filesystem
  // api documentation. note that most function names in our wrapper
  // library do not carry forward the original module's 'f_' prefix.
  
  // visit: http://elm-chan.org/fsw/ff/en/open.html for detail about
  // the file access (fa_write) and result (fr_ok) codes used here.

  dir_list();
  Serial.println();

  if (FatFs.open("test2.txt", FA_CREATE_ALWAYS | FA_READ | FA_WRITE) == FR_OK) {
    file_read_and_print_to_serial("test2.txt");
    Serial.println();
    
    FatFs.lseek(0);
    file_write("Hello. I am Inigo Montoya. You Killed my father. Prepare to die. ", "test2.txt");
    FatFs.sync();
    Serial.println();
    
    file_read_and_print_to_serial("test2.txt");
    Serial.println();
    
    // insert your code to modify the file here
    
    FatFs.close();
  }
  dir_list();
  
  while(1);
}
